import {
  ConstructorType,
  DecoratorType,
  DecoratorParametersType,
  PrimitiveType,
  PropertyKeyType,
} from './type';


export class TypeUtil {
  /**
   * Get the provide constructor.
   * @param targetClass
   */
  public static getContructor(targetClass: any): ConstructorType {
    return typeof targetClass === 'function'
        ? targetClass
        : targetClass.constructor;
  }

  /**
   * Get the provide constructor if target is an instance.
   * @param target
   * @returns {*}
   */
  public static getClass(target: any): any {
      return target.prototype ? target : target.constructor;
  }

  /**
   *
   * @param target
   * @returns {symbol}
   */
  public static getClassOrSymbol(target: any): any {
      return typeof target === 'symbol' ? target : TypeUtil.getClass(target);
  }

  /**
   * Return true if the given obj is a primitive.
   * @param target
   * @returns {boolean}
   */
  public static isPrimitiveOrPrimitiveClass(target: any): boolean {
    return TypeUtil.isString(target)
        || TypeUtil.isNumber(target)
        || TypeUtil.isBoolean(target);
  }

  /**
   *
   * @param target
   * @returns {PrimitiveType}
   */
  public static primitiveOf(target: any): PrimitiveType {
    if (TypeUtil.isString(target)) {
        return PrimitiveType.STRING;
    }
    if (TypeUtil.isNumber(target)) {
        return PrimitiveType.NUMBER;
    }
    if (TypeUtil.isBoolean(target)) {
        return PrimitiveType.BOOLEAN;
    }
    return PrimitiveType.ANY;
  }

  /**
   *
   * @param target
   * @returns {boolean}
   */
  public static isString(target: any): boolean {
    return typeof target === 'string' || target instanceof String || target === String;
  }

  /**
   *
   * @param target
   * @returns {boolean}
   */
  public static isNumber(target: any): boolean {
    return typeof target === 'number' || target instanceof Number || target === Number;
  }

  /**
   *
   * @param target
   * @returns {boolean}
   */
  public static isBoolean(target: any): boolean {
    return typeof target === 'boolean' || target instanceof Boolean || target === Boolean;
  }

  /**
   *
   * @param target
   * @returns {Boolean}
   */
  public static isArray(target: any): boolean {
    return Array.isArray(target);
  }

  /**
   * Return true if the clazz is an array.
   * @param target
   * @returns {boolean}
   */
  public static isArrayOrArrayClass(target: any): boolean {
    if (target === Array) {
        return true;
    }
    return TypeUtil.isArray(target);
  }

  /**
   * Return true if the target.
   * @param target
   * @returns {boolean}
   */
  public static isCollection(target: any): boolean {
    return TypeUtil.isArrayOrArrayClass(target)
        || target === Map
        || target instanceof Map
        || target === Set
        || target instanceof Set
        || target === WeakMap
        || target instanceof WeakMap
        || target === WeakSet
        || target instanceof WeakSet;
  }

  /**
   *
   * @param target
   * @returns {boolean}
   */
  public static isDate(target: any): boolean {
    return target === Date || target instanceof Date;
  }

  /**
   *
   * @param target
   * @returns {boolean}
   */
  public static isMethod(target: any, propertyKey: PropertyKeyType): boolean {
    if (typeof(target[propertyKey]) === undefined) {
      return false;
    }
    return typeof target[propertyKey] === 'function';
  }

  /**
   *
   * @param target
   * @returns {boolean}
   */
  public static isObject(target: any): boolean {
    return target === Object;
  }

  /**
   *
   * @param target
   * @returns {boolean}
   */
  public static isClass(target: any): boolean {
    return !TypeUtil.isPrimitiveOrPrimitiveClass(target)
        && !TypeUtil.isObject(target)
        && !TypeUtil.isDate(target)
        && target !== undefined
        && !TypeUtil.isPromise(target);
  }

  /**
   * Return true if the value is an empty string, null or undefined.
   * @param value
   * @returns {boolean}
   */
  public static isEmpty(value: any): boolean {
    return value === '' || value === null || value === undefined;
  }

  /**
   * Get object name
   */
  public static nameOf(obj: any): string {
    switch (typeof obj) {
        default:
            return '' + obj;
        case 'symbol':
            return TypeUtil.nameOfSymbol(obj);
        case 'function':
            return TypeUtil.nameOfClass(obj);
    }
  }

  /**
   * Get the provide name.
   * @param targetClass
   */
  public static nameOfClass(targetClass: any): string {
    return typeof targetClass === 'function'
        ? targetClass.name
        : targetClass.constructor.name;
  }
  /**
   * Get symbol name.
   * @param sym
   */
  public static nameOfSymbol(sym: symbol): string {
    return sym.toString().replace('Symbol(', '').replace(')', '');
  }
  /**
   *
   * @param out
   * @param obj
   * @param {{[p: string]: (collection: any[], value: any) => any}} reducers
   * @returns {any}
   */
  public static deepExtends(out: any, obj: any, reducers: { [key: string]: (collection: any[], value: any) => any } = {}): any {

    if (obj === undefined || obj === null) {
        return obj;
    }

    if (TypeUtil.isPrimitiveOrPrimitiveClass(obj) || typeof obj === 'symbol' || typeof obj === 'function') {
        return obj;
    }

    if (TypeUtil.isArrayOrArrayClass(obj)) {
        out = out || [];
    } else {
        out = out || {};
    }

    const defaultReducer = reducers.default ? reducers.default : (collection: any[], value: any) => {
        collection.push(value);
        return collection;
    };
    const set = (key: string | number, value: any) => {
        if (TypeUtil.isArrayOrArrayClass(obj)) {
            out.push(value);
        } else {
            out[key] = value;
        }
    };

    Object.keys(obj).forEach(key => {
        let value = obj[key];

        if (value === undefined || value === null) {
            return;
        }

        if (value === '' && out[key] !== '') {
            return;
        }

        if (TypeUtil.isPrimitiveOrPrimitiveClass(value) || typeof value === 'function') {
            set(key, value);
            return;
        }

        if (TypeUtil.isArrayOrArrayClass(value)) {

            value = value.map((v: any) => TypeUtil.deepExtends(undefined, v));

            set(key, []
                .concat(out[key] || [], value)
                .reduce((collection: any[], v: any) =>
                        reducers[key] ? reducers[key](collection, v) : defaultReducer(collection, v),
                        []));
            return;
        }

        // Object
        if (TypeUtil.isArrayOrArrayClass(obj)) {
            set(key, TypeUtil.deepExtends(undefined, value, reducers));
        } else {
            set(key, TypeUtil.deepExtends(out[key], value, reducers));
        }
    });

    if (TypeUtil.isArrayOrArrayClass(out)) {
        out.reduce((collection: any[], value: any) => defaultReducer(collection, value), []);
    }

    return out;
  }

  /**
   *
   * @param target
   * @returns {boolean}
   */
  public static isPromise(target: any): boolean {
    return target === Promise || target instanceof Promise;
  }

  /**
   *
   * @param target
   * @returns {any}
   */
  public static getInheritedClass(target: any): any {
    return Object.getPrototypeOf(target);
  }

  /**
   *
   * @param {any[]} args
   * @returns {DecoratorType}
   */
  public static getDecoratorType(args: any[]): DecoratorType {
    const [, propertyKey, descriptor] = args;

    if (typeof descriptor === 'number') {
        return DecoratorType.PARAMETER;
    }

    if (propertyKey && descriptor === undefined || descriptor && (descriptor.get || descriptor.set)) {
        return DecoratorType.PROPERTY;
    }
    return (descriptor && descriptor.value) ? DecoratorType.METHOD : DecoratorType.CLASS;
  }

  /**
   *
   * @param target
   * @param {PropertyKeyType} propertyKey
   * @returns {PropertyDescriptor}
   */
  public static descriptorOf(target: any, propertyKey: PropertyKeyType): PropertyDescriptor {
    return Object.getOwnPropertyDescriptor(target && target.prototype || target, propertyKey)!;
  }

  /**
   *
   * @param target
   * @param {string} propertyKey
   * @returns {DecoratorParametersType}
   */
  public static decoratorArgs(target: any, propertyKey: string): DecoratorParametersType {
    return [
        target,
        propertyKey,
        TypeUtil.descriptorOf(target, propertyKey)!,
    ];
  }

  /**
   *
   * @param target
   * @returns {Array}
   */
  public static ancestorsOf(target: any): any[] {
    const classes = [];

    let currentTarget = TypeUtil.getClass(target);

    while (TypeUtil.nameOf(currentTarget) !== '') {
        classes.unshift(currentTarget);
        currentTarget = TypeUtil.getInheritedClass(currentTarget);
    }

    return classes;
  }

  /**
   *
   * @param target
   * @param {string} name
   * @param {Function} callback
   */
  public static applyBefore(target: any, name: string, callback: Function): void {
    const original = target[name];
    target[name] = function (...args: any[]): any {
        callback(...args);
        return original.apply(this, args);
    };
  }

  /**
   *
   * @param {Promise<any>} promise
   * @param {number} time
   * @returns {Promise<any>}
   */
  public static promiseTimeout(promise: Promise<any>, time: number = 1000): Promise<{ ok: boolean, response: any }> {
    const timeout = (p: Promise<any>, t: number) => new Promise((resolve) => {
        p.then((response) => {
            resolve();
            return response;
        });
        setTimeout(() => resolve({ok: false}), t);
    });

    promise = promise.then((response) => ({ok: true, response}));

    return Promise.race([
        promise,
        timeout(promise, time),
    ]);
  }

}
